import time
import inspect
import logging
from types import GeneratorType

from pbnj.connection import Connection
from pbnj.models import Message, Command, _builtin_command
from pbnj import __version__

log = logging.getLogger("pbnj")


class Bot:
    def __init__(
        self,
        nick,
        initial_channels=[],
        username=None,
        realname=None,
        use_builtin=True,
        builtin_prefix="^\.",
        connect_wait=0,
        follow_invite=True,
        ssl=False,
    ):
        self.nick = nick
        self.username = username or nick
        self.realname = realname or nick
        self.channels = initial_channels
        self.max_msg_len = 300
        self.commands = []
        self.conn = None
        self.use_builtin = use_builtin
        self.builtin_prefix = builtin_prefix
        self.connect_wait = connect_wait
        self.follow_invite = follow_invite
        self.ssl = ssl

    def __str__(self):
        return "pbnj.Bot {}".format(self.nick)

    def __repr__(self):
        return "pbnj.Bot {}, (user: {} real: {})".format(
            self.nick, self.username, self.realname
        )

    def _is_connected(self):
        return self.conn is not None

    def _channelify(self, text_stream):
        """ensure an iterable of channels start with #"""
        for ch_name in text_stream:
            if ch_name.startswith("#"):
                yield ch_name
            else:
                yield "#" + ch_name

    def _messageify(self, text_stream):
        """turn the raw text coming off the socket into a stream of objects"""
        for raw_message in text_stream:
            yield Message(raw_message)

    def _enable_builtin_commands(self):
        """check for _builtin_command decorated commands, insert them"""
        log.debug("Checking methods for builtin commands")
        for m_name, method in inspect.getmembers(self, inspect.ismethod):
            if "_command" in dir(method):
                log.debug("{} is a command!".format(m_name))
                self.commands.append(
                    Command(self.builtin_prefix + method._filterspec, method)
                )
        log.debug(str(self.commands))

    def connect(self, addr, port=6667):
        """create a connection to an address, or return one if it already exists"""
        if not self._is_connected():
            self.conn = Connection(addr, port, __version__, use_ssl=self.ssl)
        return self.conn

    def command(self, filterspec):
        """the decorator which marks an external function as a Command in the
        bot's context
        """

        def real_decorator(function):
            log.debug("Creating command for function {}".format(function))
            c = Command(filterspec, function)
            self.commands.append(c)
            log.debug("Added to self.commands")
            return function

        log.debug("Exiting command() decorator")
        return real_decorator

    def joinall(self, channels):
        """joins a bunch of channels"""
        success = True
        for channel in self._channelify(channels):
            self.channels.append(channel)
            success = success and self.conn.join(channel)
        return success

    def part(self, channels):
        """leaves a bunch of channels :( """
        for channel in self._channelify(channels):
            self.channels.remove(channel)
            self.conn.part(channel)

    def raw_send(self, message):
        """deliver a message directly to the connection- useful for doing things
        like MODE"""
        return self.conn.send(message)

    def run(self):
        """set up and connect the bot, start looping!"""
        invitation = "INVITE " + self.nick
        if self.use_builtin:
            self._enable_builtin_commands()
        # start the connection
        with self.conn:
            # make sure we're registered to the irc network
            self.conn.register(self.username, self.nick, self.conn.addr, self.realname)
            time.sleep(self.connect_wait)
            # handle any channels the user asked us to join
            if self.channels:
                log.info("Joining initial channels")
                for channel in self.channels:
                    self.conn.join(channel)
            for msg in self._messageify(self.conn.recieve()):
                if self.follow_invite and invitation in msg.raw_msg:
                    destination = msg.raw_msg.split(":")[-1]
                    self.joinall([destination])
                self.handle(msg)

    def handle(self, message):
        """Iterates through the registered commands and attempts to find a
        command which matches the incoming Message. Does this by calling
        command.match() for each.
        """
        for command in self.commands:
            log.debug("Checking command {}".format(command.name))
            if command.match(message):  # the call
                log.info("{} matched!".format(command.name))
                resp = command(message)
                log.info("Called command method {0}".format(command.name))
                if resp:
                    log.info("Got a reply: {}".format(resp))
                    # we have something to hand back
                    if type(resp) == str:
                        log.debug("Response is a string, sending...")
                        return self.conn.message(message.reply_dest, resp)
                    elif isinstance(resp, GeneratorType):
                        log.debug("Response is a generator, giving back the contents")
                        success = True
                        for reply in resp:
                            success = success and self.conn.message(
                                message.reply_dest, reply
                            )
                        return success
                    elif isinstance(resp, bool):
                        log.debug("The function handed back a boolean, returning it")
                        return resp
                    else:
                        log.warning(
                            "Got back a weird type from command {}".format(command.name)
                        )
                        log.warning(resp)
                        return False
                break  # don't check any more methods
            else:
                log.debug("{0} failed to match {1}".format(command.name, message))
        log.debug("No matches found.")
        return False  # couldn't find a match for the command at all

    @_builtin_command("version")
    def version(self, message):
        """display the library version"""
        return "{}: {} version {}".format(message.nick, self.nick, __version__)

    @_builtin_command("ping")
    def ping(self, message):
        """pong"""
        return "{}: pong".format(message.nick)

    @_builtin_command("join")
    def join(self, message):
        """join a number of channels"""
        if len(message.args) < 1:
            return "Usage: {}join #channelname".format(self.builtin_prefix)
        else:
            log.debug("We got channels: {}".format(message.args))
            return self.joinall(message.args)

    @_builtin_command("help")
    def help(self, message):
        """return name and description of all commands"""
        if len(message.args) == 2:
            # we want detailed help for a specific command
            for command in self.commands:
                if command.name == message.args[1]:
                    return "{}: {} - {}".format(
                        message.nick, command.name, str(command)
                    )
        else:
            return "Available commands: " + " ".join(c.name for c in self.commands)



import ssl
import socket
import logging

log = logging.getLogger("pbnj")


class Connection:
    """sets up the bots connection on a socket level.
    The only *magic* this class does is respond to PING messages with PONG
    messages, which I see as an essential part of maintaining a Connection
    """

    def __init__(
        self, addr, port, version="-1", timeout=10.0, recv_bufsz=4096, use_ssl=False
    ):
        conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if use_ssl or port == 6697:
            self.conn = ssl.wrap_socket(conn, cert_reqs=ssl.CERT_REQUIRED)
        else:
            self.conn = conn
        self.ssl = use_ssl
        self.addr = addr
        self.port = port
        self.version = version
        self.conn.settimeout(timeout)
        self.conn.setblocking(1)  # will block
        self.recv_bufsz = recv_bufsz
        self.read = b""
        self.linesep = b"\r\n"
        self._connected = False

    def __str__(self):
        return "pbnj.connection.Connection to {}:{}, ssl {}".format(
            addr, port, "on" if self.ssl else "off"
        )

    def __repr___(self):
        return self.__str__()

    def __enter__(self):
        self._connect()
        return self

    def __exit__(self, type, value, traceback):
        if self._connected:
            self._cleanup()

    def _cleanup(self):
        """close down everything that needs to be closed"""
        self.send("QUIT :{0}/{1}".format(self.nick, self.version))
        log.warning("Closing socket and IRC connections")
        self._connected = False
        self.conn.close()

    def _connect(self):
        """set up the socket connection and be ready for sending data"""
        # TODO: add retry logic
        try:
            self.conn.connect((self.addr, self.port))
            self._connected = True
        except ssl.SSLError:
            log.fatal("Failed to verify the connection to the server via SSL")
            raise

    def _is_termination_message(self, message):
        return "Closing link" in message or "Server going down" in message

    def _recv(self):
        """recieve only one line from the socket"""
        while self.linesep not in self.read:
            self.read += self.conn.recv(self.recv_bufsz)
        format_msg = lambda x: (str(x[0], "utf-8"), x[1])
        message, read = format_msg(self.read.split(self.linesep, 1))
        if "\x01" in message:
            message = message.replace("\x01", "")
        log.info("RECV {0}".format(message))
        if not message:
            log.warning("Got a null message from server")
            raise ConnectionException("Null message")
        elif self._is_termination_message(message):
            log.warning("Got a termination message from server")
            raise ConnectionException("Got a termination message")
        else:
            self.read = read
            return message

    def recieve(self):
        """recieve lines of text from our socket and return them as a Generator
        """
        connected = True
        while connected:
            try:
                message = self._recv()
                if message.startswith("PING"):
                    log.debug("Replying with PONG...")
                    self.send("PONG" + message[4:])
                else:
                    yield message
            except ConnectionException:
                connected = False
        self._cleanup()

    def send(self, message):
        """helper method to convert the string, tack on a \r\n and log it"""
        try:
            self.conn.send(message.encode() + b"\r\n")
            log.info("SEND " + message)
            return True
        except Exception as e:
            log.error("Hit an exception while trying to send {}".format(message))
            return False

    def part(self, channel):
        return self.send("PART {}".format(channel))

    def join(self, channel):
        if self._connected:
            return self.send("JOIN {}".format(channel))
        else:
            log.info(
                "Tried to join a channel before starting, will join when connected"
            )
            return False

    def message(self, channel, message):
        return self.send("PRIVMSG {0} :{1}".format(channel, message))

    def register(self, user, nick, hostname, realname=None):
        """send the needed info to the IRC server after connecting"""
        realname = user if not realname else realname
        self.nick = nick
        log.info(
            "Registering on network {0} with ({1}/{2}/{3})".format(
                self.addr, user, realname, hostname
            )
        )
        n = self.send("NICK {0}".format(nick))
        return n and self.send("USER {0} {0} {2} :{1}".format(user, realname, hostname))


class ConnectionException(Exception):
    pass



"""
       |          _)
 __ \  __ \  __ \  |
 |   | |   | |   | |
 .__/ _.__/ _|  _| |
_|             ___/

"""
import sys

__version__ = "0.3.3"
__author__ = "James Luck <me@jamesluck.com>"
__title__ = "pbnj"
__license__ = "GNU GPLv3"


def default_argparser(arguments=sys.argv[1:], docstring=None, override=None):
    """this is a default argument parser for quickly creating a configured bot
    override: a pbnj.bot.Bot object that will be configured with these arguments
    docstring: override the docstring of the argument parser
    arguments: override the args being parsed
    """
    import argparse
    import logging
    from pbnj.logger import ColorFormatter

    log = logging.getLogger("pbnj")
    color_formatter = ColorFormatter()
    sh = logging.StreamHandler()
    sh.setFormatter(color_formatter)
    log.addHandler(sh)
    p = argparse.ArgumentParser(description=docstring)
    p.add_argument("-n", "--nick", default=__title__, help="specify nickname to use")
    p.add_argument(
        "-d", "--debug", action="store_true", help="increase logging verbosity to DEBUG"
    )
    p.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="decrease logging verbosity to WARNING",
    )
    p.add_argument(
        "--no-color", action="store_true", help="disable coloration of logging output"
    )
    p.add_argument(
        "--network", default="127.0.0.1", help="FQDN of IRC network to connect to"
    )
    p.add_argument(
        "--port",
        type=int,
        default=6667,
        help="specify different port for the connection",
    )
    p.add_argument(
        "-s",
        "--ssl",
        action="store_true",
        help="use SSL in your connection to the network",
    )
    p.add_argument(
        "-c",
        "--channels",
        default="",
        help="comma-separated channels to connect to when joining",
    )
    p.add_argument(
        "--user-name",
        dest="username",
        default=__title__,
        help="specify different name to use",
    )
    p.add_argument(
        "--real-name",
        dest="realname",
        default=__title__,
        help="specify different realname to use",
    )
    p.add_argument(
        "-v", "--version", help="show version of this program", action="store_true"
    )
    args = p.parse_args(arguments)
    log_lvl = (
        logging.DEBUG if args.debug else logging.WARNING if args.quiet else logging.INFO
    )
    color_formatter.color_enabled = not args.no_color
    log.setLevel(log_lvl)
    if args.version:
        print(__title__ + " version " + __version__)
        return True
    if override:
        override.nick = args.nick
        override.username = args.username
        override.realname = args.realname
        override.ssl = args.ssl
        override.connect(args.network, args.port)
        override.channels = override._channelify(args.channels.split(","))
    return args



"""a custom logging formatter that colors the messages based on the severity
of the log level. Created to help my eyes whilst debugging"""
import logging

FORMAT = "%(asctime)s %(levelname)s [%(module)s] %(message)s"


class Colors:
    esc = "\033"
    black = esc + "[0;30m"
    red = esc + "[0;31m"
    green = esc + "[0;32m"
    yellow = esc + "[0;33m"
    blue = esc + "[0;34m"
    purple = esc + "[0;35m"
    cyan = esc + "[0;36m"
    white = esc + "[0;37m"
    reset = esc + "[0m"
    level_map = {
        "WARNING": yellow,
        "INFO": green,
        "DEBUG": white,
        "CRITICAL": red,
        "ERROR": red,
    }


class ColorFormatter(logging.Formatter):
    def __init__(self, msg=FORMAT, color_enabled=True):
        logging.Formatter.__init__(self, msg)
        self.color_enabled = color_enabled

    def format(self, record):
        level = record.levelname
        if self.color_enabled and level in Colors.level_map:
            color_msg = Colors.level_map[level] + record.msg + Colors.reset
            record.msg = color_msg
        return logging.Formatter.format(self, record)



import re
import logging

log = logging.getLogger("pbnj")

attr_filter = lambda x: {
    a: getattr(x, a) for a in dir(x) if not "__" in a and not callable(getattr(x, a))
}


class Message:
    """object to get passed between Bot class and its command methods. Separates
    the logic of parsing a string from the socket into more actionable objects
    """

    def __init__(self, raw_msg):
        self.raw_msg = raw_msg
        self.message = None
        self.parse()

    def __str__(self):
        return self.message or self.raw_msg

    def __repr__(self):
        return self.raw_msg

    def __eq__(self, other):
        self_attr = attr_filter(self)
        for item, val in self_attr.items():
            if val != getattr(other, item, False):
                return False
        return True

    def parse(self):
        """ Parse the message
        TODO handle all the numeric ones
        TODO remove dependence on regular expressions
        """
        sp = self.raw_msg.split()
        host = sp[0]
        if "@" in host:
            hostmask = host[1:] if host.startswith(":") else host
            re_matches = re.match(
                "^([a-zA-Z0-9_\-]+)!~?([a-zA-Z0-9\ ]+)@(.*)", hostmask
            )
            try:
                re_groups = re_matches.groups()
                self.nick = re_groups[0]
                self.realname = re_groups[1]
                self.host = re_groups[2]
            except Exception as e:
                log.error("The regex fucked up! On this input for the hostmask")
                log.error(hostmask)
                raise
            self.type = sp[1]
        else:
            # this is a server directly sending us something
            self.host = host
            code = sp[1]
            self.type = int(code) if code.isdigit() else code
        if self.type == "PRIVMSG":
            self.dest = sp[2]
            if "#" in self.dest:
                self.reply_dest = self.dest
            else:
                self.reply_dest = self.nick
            m = " ".join(sp[3:])
            msg = m[1:] if m.startswith(":") else m
            if msg.startswith("ACTION"):
                self.type = "ACTION"
                msg = msg[7:]  # actions are privmsgs, why
            self.message = msg
            self.args = msg.split()[1:]


class _builtin_command:
    """lightweight decorator for doing the marking of builtin commands as such
    before runtime. This is necessarily outside of the Bot class because this
    needs to mark methods of that class as loadable before the class' __init__
    is called"""

    def __init__(self, filterspec):
        self.filterspec = filterspec

    def __call__(self, f):
        def wrapped_f(*args):
            return f(*args)

        wrapped_f._command = True
        wrapped_f._filterspec = self.filterspec
        wrapped_f.__name__ = f.__name__
        wrapped_f.__doc__ = f.__doc__
        return wrapped_f


class Command:
    """holds the message filtering specification of a command registered with
    the bot (a callable or string regex) as well as the callback to hit if a
    message is matched"""

    def __init__(self, filterspec, callback):
        if not (callable(filterspec) or isinstance(filterspec, str)):
            raise ValueError(
                "filterspec arg for Command classes must be callable or a string (valid regex)!"
            )
        self.filterspec = filterspec
        self.callback = callback
        self.name = callback.__name__
        self.__doc__ = callback.__doc__

    def __str__(self):
        return self.__doc__ or "This triggers it: {}".format(self.filterspec)

    def __repr__(self):
        return self.name

    def __call__(self, *args):
        return self.callback(*args)

    def __eq__(self, other):
        return (
            self.filterspec == other.filterspec
            and self.callback == other.callback
            and self.name == other.name
            and self.__doc__ == other.__doc__
        )

    def match(self, message):
        """try to match an incoming message (quickly) return something falsey if
        not"""
        if callable(self.filterspec):
            return self.filterspec(message)
        else:
            """we only support the naive regex format for privmsg types"""
            if message.type == "PRIVMSG":
                log.debug(
                    "Trying to match {} with {}".format(
                        message.message, self.filterspec
                    )
                )
                return re.match(self.filterspec, message.message)
            return None



#!/usr/bin/env python
from setuptools import setup
import sys

sys.path.extend('.')
from pbnj import __version__

test_requirements = ['pytest>=2.8.0', 'pytest-cov']

setup(
    name = 'pbnj',
    packages = ['pbnj'],
    version = __version__,
    description = 'an IRC bot library and framework, focused on simplicity and portability',
    author = 'James Luck',
    author_email = 'me@jamesluck.com',
    license='GNU GPLv3',
    url = 'https://github.com/delucks/pbnj',
    download_url = 'https://github.com/delucks/pbnj/tarball/v{}'.format(__version__),
    tests_requirements=test_requirements,
    keywords = ['irc', 'chatbot', 'bot', 'portable', 'framework'],
    classifiers = [],
)
